/****************************************************************************/
/*                                                                          */
/*                   Copyright 2003 University of Glasgow                   */
/*                                                                          */
/*                           All rights reserved.                           */
/*                                                                          */
/****************************************************************************/

/****************************************************************************/
/* FILE          : Checker.java                                             */
/* DESCRIPTION   : Typechecker for the Simple Functional Language           */
/*                                                                          */
/* AUTHOR        : S.J.Gay                                                  */
/* DATE          : 28th August 2003                                         */
/*                                                                          */
/* LAST MODIFIED : S.J.Gay                                                  */
/* DATE          : 8th October 2003                                         */
/****************************************************************************/

package sfl.checker;

import java.util.*;
import java.io.*;
import sfl.node.*;
import sfl.analysis.*;
import sfl.error.*;
import sfl.types.*;
import sfl.env.*;

public class Checker extends DepthFirstAdapter {

    ErrorTable errorTable = new ErrorTable();
    Env env = new Env();
    boolean mutual = false;

    public ErrorTable getErrorTable() {
	return errorTable;
    }

    public void setEnv(Env e) {
	env = e;
    }

    public void allowMutualRecursion() {
	mutual = true;
    }

    public void outStart(Start node) {
	setOut(node, getOut(node.getPProgram()));
    }

    public void outAProgram(AProgram node) {
	setOut(node, getOut(node.getExpression()));
    }

    public void outATermExpression(ATermExpression node)
    {
        setOut(node, getOut(node.getTerm()));
    }

    public void outAExpTerm(AExpTerm node)
    {
	setOut(node, getOut(node.getExpression()));
    }

    public void outABoolLitTerm(ABoolLitTerm node)
    {
	setOut(node, new BoolType());
    }

    public void outAIntLitTerm(AIntLitTerm node)
    {
	setOut(node, new IntType());
    }

    public void outAIdTerm(AIdTerm node) {
	EnvEntry entry = env.get(node.getId().getText());
	if (entry == null) {
	    errorTable.add(node.getId().getLine(),
			   new UndefinedIdError(node.getId().getText()));
	}
	else if (entry instanceof FunEntry) {
	    errorTable.add(node.getId().getLine(),
			   new FunNoArgsError(node.getId().getText()));
	}
	else {
	    setOut(node, ((VarEntry)entry).getType());
	};
    }

    public void outAAppTerm(AAppTerm node) {
	// Look for the function name in the environment
	EnvEntry entry = env.get(node.getId().getText());
	if (entry == null) {
	    // The function name is not in the environment
	    errorTable.add(node.getId().getLine(),
			   new UndefinedFunError(node.getId().getText()));
	}
	else if (entry instanceof VarEntry) {
	    // The function name is there but refers to a variable
	    errorTable.add(node.getId().getLine(),
			   new AppVarError(node.getId().getText()));
	}
	else {
	    // First count the arguments and find out their types
	    int actualArity;
	    Type[] actualArgTypes;
	    if (node.getArgs() == null) {
		// There are no arguments
		actualArity = 0;
		actualArgTypes = new Type[0];
	    }
	    else {
		// First arg and remaining args are stored separately
		Object[] actualArgsTail = 
		    ((AArgs)node.getArgs()).getCommaExpression().toArray();
		// Count the arguments
		actualArity = actualArgsTail.length + 1;
		actualArgTypes = new Type[actualArity];
		// Copy the first argument type into the array
		actualArgTypes[0] = 
		    (Type)getOut(((AArgs)node.getArgs()).getExpression());
		// Copy the rest of the argument types
		for (int i = 1; i < actualArity; i++) {
		    actualArgTypes[i] = 
			(Type)getOut(((ACommaExpression)actualArgsTail[i-1])
				     .getExpression());
		};
	    };
	    int formalArity = ((FunEntry)entry).getArity();
	    if (formalArity != actualArity) {
		errorTable.add(node.getId().getLine(),
			       new WrongNumArgsError(node.getId().getText(),
						     formalArity,actualArity));
	    };
	    // Even if the wrong number of arguments have been supplied,
	    // check the types of as many as we can
	    for (int i = 0; i < formalArity; i++) {
		Type formalArgType = ((FunEntry)entry).getArgType(i);
		if (i < actualArity) {
		    if (actualArgTypes[i] != null) {
			// if actualArgTypes[i] == null then there was a
			// type error in that argument so we don't need
			// to produce another error
			if (!(formalArgType.equal(actualArgTypes[i]))) {
			    errorTable.add(node.getId().getLine(),
				new WrongArgTypeError(node.getId().getText(),i,
						 formalArgType.name(),
						 actualArgTypes[i].name()));
			};
		    };
		};
	    };
	    setOut(node,((FunEntry)entry).getResultType());
	};
    }

    public void outAPlusExpression(APlusExpression node)
    {
	Type leftType = (Type)getOut(node.getLeft());
	Type rightType = (Type)getOut(node.getRight());

	if (leftType != null) {
	    if (!(leftType instanceof IntType)) {
		errorTable.add(node.getPlus().getLine(),
			       new PlusLeftError(leftType.name()));
	    }
	};

	if (rightType != null) {
	    if (!(rightType instanceof IntType)) {
		errorTable.add(node.getPlus().getLine(),
			       new PlusRightError(rightType.name()));
	    }
	};

	if ((leftType instanceof IntType) 
	    && (rightType instanceof IntType)) {
	    setOut(node, new IntType());
	};
    }

    public void outAAndExpression(AAndExpression node)
    {
	Type leftType = (Type)getOut(node.getLeft());
	Type rightType = (Type)getOut(node.getRight());

	if (leftType != null) {
	    if (!(leftType instanceof BoolType)) {
		errorTable.add(node.getAnd().getLine(),
			       new AndLeftError(leftType.name()));
	    }
	};

	if (rightType != null) {
	    if (!(rightType instanceof BoolType)) {
		errorTable.add(node.getAnd().getLine(),
			       new AndRightError(rightType.name()));
	    }
	};

	if ((leftType instanceof BoolType) 
	    && (rightType instanceof BoolType)) {
	    setOut(node, new BoolType());
	};
    }

    public void outAEqExpression(AEqExpression node)
    {
	Type leftType = (Type)getOut(node.getLeft());
	Type rightType = (Type)getOut(node.getRight());

	if (leftType != null) {
	    if (!(leftType instanceof IntType)) {
		errorTable.add(node.getEq().getLine(),
			       new EqLeftError(leftType.name()));
	    }
	};

	if (rightType != null) {
	    if (!(rightType instanceof IntType)) {
		errorTable.add(node.getEq().getLine(),
			       new EqRightError(rightType.name()));
	    }
	};

	if ((leftType instanceof IntType) 
	    && (rightType instanceof IntType)) {
	    setOut(node, new BoolType());
	};
    }

    public void outACondExpression(ACondExpression node)
    {
	Type condType = (Type)getOut(node.getCond());
	Type thenType = (Type)getOut(node.getThenBranch());
	Type elseType = (Type)getOut(node.getElseBranch());

	if (condType != null) {
	    if (!(condType instanceof BoolType)) {
		errorTable.add(node.getIf().getLine(),
			       new IfCondError(condType.name()));
	    }
	};

	if ((thenType != null) && (elseType != null)) {
	    if (!(thenType.equal(elseType))) {
		errorTable.add(node.getIf().getLine(),
		    new IfMismatchError(thenType.name(),elseType.name()));
	    }
	    else {
		if (condType instanceof BoolType) {
		    setOut(node,thenType);
		}
	    }
	};

    }

    public void outAIntTy(AIntTy node) {
	setOut(node, new IntType());
    }

    public void outABoolTy(ABoolTy node) {
	setOut(node, new BoolType());
    }

    public void caseAFunDef(AFunDef node) {
	// This is the only kind of node for which we override the "case"
	// method. This is because we will only visit the body of the
	// function if the formal arguments make sense.

	// Visit all of the typed arguments in order to convert the types
	// from the syntax tree representation (ATy) to our representation
	// (Type)
	node.getTy().apply(this);
        if (node.getTypedArgs() != null) {
            node.getTypedArgs().apply(this);
        };
	String fName = node.getId().getText();
	int arity;
	Type[] argTypes;
	String[] argNames;
	ATypedArgs typedArgs = (ATypedArgs)node.getTypedArgs();
	// Construct arrays of argument names and types, in both cases:
	// no arguments, or some arguments.
	if (typedArgs == null) {
	    arity = 0;
	    argTypes = new Type[0];
	    argNames = new String[0];
	}
	else {
	    Object[] typedArgsTail = typedArgs.getCommaTypedArg().toArray();
	    arity = typedArgsTail.length + 1;
	    argTypes = new Type[arity];
	    argNames = new String[arity];
	    argTypes[0] = 
		(Type)getOut(((ATypedArg)typedArgs.getTypedArg()).getTy());
	    argNames[0] = 
		((ATypedArg)typedArgs.getTypedArg()).getId().getText();
	    for (int i = 1; i < arity; i++) {
		ACommaTypedArg cta = (ACommaTypedArg)typedArgsTail[i-1];
		ATypedArg ta = (ATypedArg)cta.getTypedArg();
		argTypes[i] = (Type)getOut(ta.getTy());
		argNames[i] = ta.getId().getText();
	    };
	};
	Type resultType = (Type)getOut(node.getTy());
	FunEntry fEntry = new FunEntry(resultType,arity,argNames,argTypes);
	// Add the function declaration (type information) to the environment.
	// Doing it here means that it is available while checking the
	// function body: recursive functions are supported.
	// If mutual recursion is allowed then the function declaration is
	// already in the environment from the first pass.
	if ((env.put(fName,fEntry) != null) & (!mutual)) {
	    errorTable.add(node.getId().getLine(),
			   new FunRedefinedError(fName));
	}
	else {
	    env.openScope();
	    boolean error = false;
	    // Add each typed argument to the environment, making sure that
	    // no name appears twice.
	    for (int i = 0; i < arity; i++) {
		if (env.put(argNames[i],new VarEntry(argTypes[i])) != null) {
		    errorTable.add(node.getId().getLine(),
				   new ArgRedefinedError(argNames[i]));
		    error = true;
		};
	    };
	    // If the arguments all have different names, then check the
	    // function body.
	    if (!error) {
		node.getExpression().apply(this);
		Type bodyType = (Type)getOut(node.getExpression());
		if (bodyType != null) {
		    if (!bodyType.equal(resultType)) {
			errorTable.add(node.getId().getLine(),
			   new WrongResultTypeError(fName,resultType.name(),
						    bodyType.name()));
		    };
		};
	    };
	    // Remove the typed arguments from the environment, ensuring that
	    // their scope is the function body and no more.
	    env.closeScope();
	}
    }

}
