/****************************************************************************/
/*                                                                          */
/*                   Copyright 2003 University of Glasgow                   */
/*                                                                          */
/*                           All rights reserved.                           */
/*                                                                          */
/****************************************************************************/

/****************************************************************************/
/* FILE          : Mutual.java                                              */
/* DESCRIPTION   : Support for mutually recursive functions in SFL          */
/*                                                                          */
/* AUTHOR        : S.J.Gay                                                  */
/* DATE          : 23rd September 2003                                      */
/*                                                                          */
/* LAST MODIFIED : S.J.Gay                                                  */
/* DATE          : 8th October 2003                                         */
/****************************************************************************/

package sfl.checker;

import java.util.*;
import java.io.*;
import sfl.node.*;
import sfl.analysis.*;
import sfl.error.*;
import sfl.types.*;
import sfl.env.*;

public class Mutual extends DepthFirstAdapter {

    ErrorTable errorTable = new ErrorTable();
    Env env = new Env();

    public ErrorTable getErrorTable() {
	return errorTable;
    }

    public Env getEnv() {
	return env;
    }

    public void outAIntTy(AIntTy node) {
	setOut(node, new IntType());
    }

    public void outABoolTy(ABoolTy node) {
	setOut(node, new BoolType());
    }

    public void caseAFunDef(AFunDef node) {
	// This is similar to caseAFunDef in Checker, but here we only
	// put the FunEntry into the environment, and don't check the
	// body. We don't check for repeated formal arguments here.

	// Visit all of the typed arguments in order to convert the types
	// from the syntax tree representation (ATy) to our representation
	// (Type)
	node.getTy().apply(this);
        if (node.getTypedArgs() != null) {
            node.getTypedArgs().apply(this);
        };
	String fName = node.getId().getText();
	int arity;
	Type[] argTypes;
	String[] argNames;
	ATypedArgs typedArgs = (ATypedArgs)node.getTypedArgs();
	// Construct arrays of argument names and types, in both cases:
	// no arguments, or some arguments.
	if (typedArgs == null) {
	    arity = 0;
	    argTypes = new Type[0];
	    argNames = new String[0];
	}
	else {
	    Object[] typedArgsTail = typedArgs.getCommaTypedArg().toArray();
	    arity = typedArgsTail.length + 1;
	    argTypes = new Type[arity];
	    argNames = new String[arity];
	    argTypes[0] = 
		(Type)getOut(((ATypedArg)typedArgs.getTypedArg()).getTy());
	    argNames[0] = 
		((ATypedArg)typedArgs.getTypedArg()).getId().getText();
	    for (int i = 1; i < arity; i++) {
		ACommaTypedArg cta = (ACommaTypedArg)typedArgsTail[i-1];
		ATypedArg ta = (ATypedArg)cta.getTypedArg();
		argTypes[i] = (Type)getOut(ta.getTy());
		argNames[i] = ta.getId().getText();
	    };
	};
	Type resultType = (Type)getOut(node.getTy());
	FunEntry fEntry = new FunEntry(resultType,arity,argNames,argTypes);
	// Add the function declaration (type information) to the environment.
	if (env.put(fName,fEntry) != null) {
	    errorTable.add(node.getId().getLine(),
		    new FunRedefinedError(fName));
	}
	else {
	    env.openScope();
	    // Add each typed argument to the environment, making sure that
	    // no name appears twice. We don't make use of this environment,
	    // because we don't check the function body, but this is a
	    // convenient way of checking that the formal arguments are all
	    // different.
	    for (int i = 0; i < arity; i++) {
		if (env.put(argNames[i],new VarEntry(argTypes[i])) != null) {
		    errorTable.add(node.getId().getLine(),
				   new ArgRedefinedError(argNames[i]));
		};
	    };
	    env.closeScope();
	}
    }

}
